import collections
import functools


def pipe(*functions):
    def closure(x):
        for fn in functions:
            if not out:
                out = fn(x)
            else:
                out = fn(out)
        return out

    return closure


class memoize(object):
    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)



from decimal import Decimal

import numpy as np

# Implementation reference:
# http://stackoverflow.com/a/29955556/1072288

LOG10_OF_2 = Decimal(
    3.010299956639811952137388947244930267681898814621085413104274611e-1)


def sigfig_round(x, n, ignore_left_of_decimal=False, eps=1e-7):

    if not (type(n) is int or np.issubdtype(n, np.integer)):
        raise TypeError("sigfig_round: n must be an integer.")

    if not np.all(np.isreal(x)):
        raise TypeError("sigfig_round: all x must be real.")

    if n <= 0:
        raise ValueError("sigfig_round: n must be positive.")

    if ignore_left_of_decimal:
        # Note: doesn't return true significant figure representation, as it's not
        # counting left of decimal sigfigs.
        left_of_decimal = np.trunc(x)
        right_of_decimal = x - left_of_decimal
        return left_of_decimal + sigfig_round(right_of_decimal, n)
    else:
        mantissas, binary_exponents = np.frexp(x)
        decimal_exponents = (LOG10_OF_2 * binary_exponents).astype(np.float64)
        int_parts = np.floor(decimal_exponents)
        mantissas *= 10.0**(decimal_exponents - int_parts)
        return np.around(mantissas.values, decimals=n - 1) * 10.0**int_parts



from six.moves import configparser as CP
from sqlalchemy.engine.url import URL
from sqlalchemy.engine import create_engine

def get_odbc_engine(name, odbc_filename='/etc/odbc.ini', database=None):
    """
    Looks up the connection details in an odbc file and returns a SQLAlchemy engine initialized with those details.
    """

    parser = CP.ConfigParser()
    parser.read(odbc_filename)

    cfg_dict = dict(parser.items(name))

    if database:
        cfg_dict['database'] = database

    connection_href = str(URL(**cfg_dict))

    engine = create_engine(connection_href)

    return engine



def sql_hist(table, column, where=''):
    """
    Queries a table, selecting the column and count of the values in that column. Generates
    a histogram plot of that data for display in a jupyter notebook.
    """

    # this depends on the type of table we're looking at, mysql, postgres, etc...
    # might want to template these queries
    sql = """
    SELECT something FROM %(table)s
    %(where)s
    """
    pass

def sql_scatter(table, x_col, y_col, coloring_dimension=None, where=''):
    """
    Queries a table for data in x_col and y_col, then displays them (optionally colored by
    coloring_dimension) onto a scatterplot.
    """
    pass



#!/usr/bin/env python

from setuptools import setup

with open('VERSION', 'r') as f:
    version = f.read()

setup(name='corgi',
      version=version,
      description='Python data and analysis tools',
      author='Paul English',
      author_email='paulnglsh@gmail.com',
      url='https://github.com/log0ymxm/corgi',
      packages=['corgi'],
      install_requires=[
          'numpy',
          'six',
          'sqlalchemy',
      ],
      )



from collections import defaultdict
from itertools import groupby


def groupdict(col, fn, value_fn=list):
    "Performs a groupby on a collection and consumes the generator result into a python dict."
    sorted_col = sorted(col, key=fn)
    return defaultdict(list, {k: value_fn(v) for k, v in groupby(sorted_col, key=fn)})



import six
from six.moves import configparser as CP
from sqlalchemy.engine.url import URL
from sqlalchemy.engine import create_engine

# TODO want to probabilistically infer what a mixed dtype column type is

def df_drop_cols(df, cols):
    return df.drop(cols, axis=1)

def df_encode_factor_cols(df, cols):
    obj_df = df[cols].astype(object)
    encode_df = pd.get_dummies(obj_df[cols])
    ret_df = df_drop_cols(df, cols)
    return pd.concat([ret_df, encode_df], axis=1)

def df_convert_dates(df, cols):
    # TODO
    # Convert dates
    X[date_cols] = X[date_cols].apply(pd.to_datetime)

    new_cols = [c + '_month' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.month)
    numeric_cols += new_cols

    new_cols = [c + '_day' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.day)
    numeric_cols += new_cols

    new_cols = [c + '_dayofyear' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.dayofyear)
    numeric_cols += new_cols

    new_cols = [c + '_week' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.week)
    numeric_cols += new_cols

    new_cols = [c + '_weekday' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.weekday)
    numeric_cols += new_cols

    new_cols = [c + '_quarter' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.quarter)
    numeric_cols += new_cols

    new_cols = [c + '_year' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.year)
    numeric_cols += new_cols

    new_cols = [c + '_is_month_start' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.is_month_start)

    new_cols = [c + '_is_month_end' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.is_month_end)

    new_cols = [c + '_is_quarter_start' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.is_quarter_start)

    new_cols = [c + '_is_quarter_end' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.is_quarter_end)

    new_cols = [c + '_is_year_start' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.is_year_start)

    new_cols = [c + '_is_year_end' for c in date_cols]
    X[new_cols] = X[date_cols].apply(lambda x: x.dt.is_year_end)

    X = X.drop(date_cols, axis=1)
    year_df = X[year_cols]
    year_df = year_df.where((pd.notnull(year_df)), None) # convert NaN to None
    year_df = year_df.apply(lambda x: pd.to_datetime(x, format='%Y', coerce=True))
    X[year_cols] = year_df
    X[year_cols] = X[year_cols].apply(lambda x: x.dt.year)
    numeric_cols += year_cols

    return "!!!! TODO"

def df_standardize_numeric(df, numeric_cols):
    # Standardize numeric columns
    num_df = X[numeric_cols].astype(np.float64)
    num_df = (num_df - num_df.mean())/num_df.std()
    ret_df = X.drop(numeric_cols, axis=1)
    X = pd.concat([ret_df, num_df], axis=1)

    return "!!!!! TODO"

def preprocess(data,
               drop_cols=None,
               factor_cols=None,
               date_cols=None,
               numeric_cols=None):
    """
    TODO
    """

    df = pd.DataFrame(data)

    df = df_drop_cols(df, drop_cols)
    df = df_encode_factor_cols(df, factor_cols)
    df = df_encode_dates(df, date_cols)
    df = df_standardize_numeric(df, numeric_cols)

    # Fill NaN values
    X = X.fillna(X.mean())

    return X

def sql_preprocess_table(table, **kwargs):
    """
    Initailize a dataframe using probabilistic data cleaning heuristics from a sql table.
    """
    # TODO df from sql table

    return magic_preprocess(df, **kwargs)

def proportion_na(df):
    pass

def duplicate_rows(df):
    pass

def constant_columns(df):
    pass

def infer_column_type(df, col):
    # hist of characters...
    # match to thingy
    pass
